/***************************************************************************
*   Copyright (C) 2003 by Hideki Ikemoto                                  *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include <kio/netaccess.h>

#include <qfile.h>
#include <qtextcodec.h>
#include <qstringlist.h>
#include <qregexp.h>
#include <klistview.h>
#include <kpopupmenu.h>
#include <klocale.h>
#include <kdebug.h>
#include <qvaluelist.h>
#include <qheader.h>
#include <qclipboard.h>
#include <kstandarddirs.h>
#include <kconfig.h>

#include "kitaboardview.h"
#include "kita.h"
#include "libkita/qcp932codec.h"
#include "libkita/favoriteboards.h"
#include "libkita/board.h"
#include "libkita/category.h"
#include "libkita/kitaconfig.h"
#include "libkita/signalcollection.h"

QCp932Codec* KitaBoardView::m_cp932Codec = NULL;

KitaBoardView::KitaBoardView( QWidget *parent, const char *name )
        : KitaBoardViewBase( parent, name )
        , m_favorites( 0 )
{
    m_boardList->setSorting( -1 );
    m_boardList->addColumn( i18n( "board name" ) );
    m_boardList->header() ->setClickEnabled( FALSE );

    Kita::SignalCollection* signalCollection = Kita::SignalCollection::getInstance();
    connect( m_boardList, SIGNAL( mouseButtonClicked( int, QListViewItem*, const QPoint&, int ) ),
             SLOT( slotMouseButtonClicked( int, QListViewItem* ) ) );
    connect( m_boardList, SIGNAL( returnPressed( QListViewItem* ) ), SLOT( loadBoard( QListViewItem* ) ) );
    connect( m_boardList, SIGNAL( contextMenuRequested( QListViewItem*, const QPoint&, int ) ),
             SLOT( slotContextMenuRequested( QListViewItem*, const QPoint&, int ) ) );
    connect( Kita::FavoriteBoards::getInstance(), SIGNAL( changed() ), SLOT( refreshFavoriteBoards() ) );
    connect( this, SIGNAL( openURLRequestExt(
                               const KURL&, const KParts::URLArgs&, QString, int, int,
                               const KURL&, const KURL&, const QString&, const QString& ) ),
             signalCollection, SIGNAL( openURLRequestExt(
                                           const KURL& , const KParts::URLArgs&, QString, int, int,
                                           const KURL&, const KURL&, const QString&, const QString& ) ) );
}

KitaBoardView::~KitaBoardView()
{
    saveOpened();
}

QString KitaBoardView::getCategory( const QString& line ) const
{
    QRegExp regexp( "<BR><BR><B>(.*)</B><BR>", false );
    if ( regexp.search( line ) != -1 ) {
        return regexp.cap( 1 );
    } else {
        return QString::null;
    }
}

QString KitaBoardView::getBoardURL( const QString& line ) const
{
    QRegExp regexp( "<A HREF=([^ ]*).*>(.*)</A>", false );
    if ( regexp.search( line ) != -1 ) {
        QString board_url = regexp.cap( 1 );
        QString board_title = regexp.cap( 2 );
        Kita::Board::setName( board_url, board_title );
        return board_url;
    } else {
        return QString::null;
    }
}

bool KitaBoardView::isBoardURL( const QString& url ) const
{
    QRegExp url_2ch( "http://.*\\.2ch\\.net/.*" );
    QRegExp url_bbspink( "http://.*\\.bbspink\\.com/.*" );
    QRegExp url_www_2ch( "http://www\\.2ch\\.net/.*" );
    QRegExp url_machibbs( "http://.*\\.machi\\.to/.*" );

    if ( url.isEmpty() ) return false;

    if ( url_2ch.search( url ) == -1 && url_bbspink.search( url ) == -1
            && url_machibbs.search( url ) == -1 ) return false;
    if ( url_www_2ch.search( url ) != -1 ) return false;

    return true;
}

void KitaBoardView::downloadBoardList()
{
    QString tmpFile;
    QString url = KitaConfig::boardListURL();
    if ( ! KIO::NetAccess::download( url, tmpFile ) ) {
        return ;
    }

    QFile file( tmpFile );
    if ( ! file.open( IO_ReadOnly ) ) {
        return ;
    }

    QTextStream stream( &file );
    if ( !m_cp932Codec ) m_cp932Codec = new QCp932Codec();
    stream.setCodec( m_cp932Codec );
    QString html = stream.read();

    QString configPath = locateLocal( "appdata", "board_list" );
    KConfig config( configPath );

    // parse
    QStringList list;
    QValueList<Kita::Category> categoryList = getCategoryList( html );
    QValueList<Kita::Category>::iterator it;

    for ( it = categoryList.begin(); it != categoryList.end(); ++it ) {
        Kita::Category category = ( *it );
        QValueList<QString> board_url_list = category.getBoardURLList();

        if ( board_url_list.isEmpty() ) {
            continue;
        }

        list << category.name();
        config.setGroup( category.name() );

        QValueList<QString>::iterator it2;
        int count = 0;
        for ( it2 = board_url_list.begin(); it2 != board_url_list.end(); ++it2 ) {
            QString key = QString( "item%1").arg( count );
            QString boardURL = *it2;
            QString boardName = Kita::Board::getName( *it2 );
            QStringList tmpList;
            tmpList << boardURL;
            tmpList << boardName;
            config.writeEntry( key, tmpList );
            count++;
        }
    }
    config.setGroup( QString::null );
    config.writeEntry( "Categories", list );
}

void KitaBoardView::updateBoardList()
{
    downloadBoardList();
    showBoardList();
}

void KitaBoardView::showBoardList()
{
    // clear list
    m_boardList->clear();
    m_favorites = NULL;

    QString configPath = locateLocal( "appdata", "board_list" );
    KConfig config( configPath );
    QStringList category_list = config.readListEntry( "Categories" );

    KListViewItem* previous = 0;
    QStringList::iterator it;
    for ( it = category_list.begin(); it != category_list.end(); ++it ) {
        QString category = ( *it );

        config.setGroup( category );
        KListViewItem* categoryItem = new KListViewItem( m_boardList, previous, category );
        KListViewItem* previousBoard = 0;

        for ( int count = 0; ; count++ ) {
            QString key = QString( "item%1" ).arg( count );
            QStringList value = config.readListEntry( key );
            if ( value.count() != 2 ) {
                break;
            }
            QString boardURL = value[0];
            QString boardName = value[1];
            Kita::Board::setName( boardURL, boardName );
            previousBoard = new KListViewItem( categoryItem, previousBoard, boardName, boardURL );
        }
        previous = categoryItem;
    }

    loadExtBoard();
    refreshFavoriteBoards();
}


/* private */
void KitaBoardView::loadExtBoard()
{
    QString configPath = locateLocal( "appdata", "extbrd.conf" );
    QFile file( configPath );
    if ( file.open( IO_ReadOnly ) ) {

        QTextStream stream( &file );
        stream.setEncoding( QTextStream::UnicodeUTF8 );

        QStringList list;
        QString str;

        KListViewItem* categoryItem = new KListViewItem( m_boardList, "extboard" );

        while ( ( str = stream.readLine() ) != QString::null ) {
            if ( ! str.isEmpty() ) {
                QStringList list = QStringList::split( "<>", str );

                if ( list.size() == 2 ) {
                    QString board_title = list[ 0 ];
                    QString board_url = list[ 1 ];

                    new KListViewItem( categoryItem, board_title, board_url );
                    Kita::Board::setName( board_url, board_title );
                }
            }
        }

        file.close();
    }
}


QValueList<Kita::Category> KitaBoardView::getCategoryList( const QString& html ) const
{
    QValueList<Kita::Category> result;

    QStringList lines = QStringList::split( "\n", html );
    QStringList::iterator it;

    Kita::Category current_category;
    for ( it = lines.begin(); it != lines.end(); ++it ) {
        QString category_name = getCategory( *it );
        if ( category_name != QString::null ) {
            if ( ! current_category.name().isEmpty() ) {
                result.append( current_category );
            }
            current_category = Kita::Category( category_name );
        } else {
            QString boardURL = getBoardURL( *it );

            if ( isBoardURL( boardURL ) ) {
                if ( ! current_category.name().isEmpty() ) {
                    current_category.append( boardURL );
                }
            }
        }
    }
    return result;
}

void KitaBoardView::refreshFavoriteBoards()
{
    if ( ! m_favorites ) {
        m_favorites = new KListViewItem( m_boardList, 0, i18n( "Favorites" ) );
    }

    do {
        QListViewItem* child = m_favorites->firstChild();
        delete child;
    } while ( m_favorites->childCount() != 0 );

    QValueList<KURL> boards = Kita::FavoriteBoards::boards();
    QValueList<KURL>::iterator it;

    for ( it = boards.begin(); it != boards.end(); ++it ) {
        QString boardURL = ( *it ).url();
        QString boardName = Kita::Board::getName( boardURL );
        new KListViewItem( m_favorites, 0, boardName, boardURL );
    }
}

void KitaBoardView::loadBoard( QListViewItem* item )
{
    if ( ! item ) return ;

    QString boardName = item->text( 0 );
    QString boardURL = item->text( 1 );

    if ( boardURL.isEmpty() ) {
        return ;
    }

    if ( KitaConfig::alwaysUseTab() ) {
        emit openURLRequestExt( boardURL, KParts::URLArgs(), "kita_open_2chboard", 1 );
    } else {
        emit openURLRequestExt( boardURL, KParts::URLArgs(), "kita_open_2chboard", 0 );
    }
}

void KitaBoardView::setFont( const QFont& font )
{
    m_boardList->setFont( font );
}

/* public slot :*/ /* virtual */
void KitaBoardView::setFocus()
{
    m_boardList->setFocus();
}


void KitaBoardView::slotContextMenuRequested( QListViewItem* item, const QPoint& point, int )
{
    enum {
        Menu_OpenWithBrowser,
        Menu_OpenWithNewTab,
        Menu_CopyURL,
        Menu_CopyTitleAndURL,
        Menu_AddToFavorites,
        Menu_RemoveFromFavorites
    };

    if ( item == 0 ) {
        return ;
    }

    KPopupMenu popup( 0 );
    popup.insertItem( i18n( "Open with Web Browser" ), Menu_OpenWithBrowser );
    // FIXME: open with "new tab" is meaningless.
    popup.insertItem( i18n( "Open with New Tab" ), Menu_OpenWithNewTab );
    popup.insertItem( i18n( "Copy URL" ), Menu_CopyURL );
    popup.insertItem( i18n( "Copy title and URL" ), Menu_CopyTitleAndURL );
    if ( item->parent() == m_favorites ) {
        popup.insertItem( i18n( "Remove from Favorites" ), Menu_RemoveFromFavorites );
    } else {
        popup.insertItem( i18n( "Add to Favorites" ), Menu_AddToFavorites );
    }

    QString boardName = item->text( 0 );
    KURL boardURL = item->text( 1 );
    QClipboard* clipboard = QApplication::clipboard();

    switch ( popup.exec( point ) ) {
    case Menu_OpenWithBrowser:
        emit openURLRequestExt( boardURL, KParts::URLArgs(), "text/html" );
        break;
    case Menu_OpenWithNewTab:
        emit openURLRequestExt( boardURL, KParts::URLArgs(), "kita_open_2chboard", 1 );
        break;
    case Menu_CopyURL:
        clipboard->setText( boardURL.prettyURL() );
        break;
    case Menu_CopyTitleAndURL:
        clipboard->setText( boardName + "\n" + boardURL.prettyURL() );
        break;
    case Menu_AddToFavorites:
        Kita::FavoriteBoards::append( boardURL );
        break;
    case Menu_RemoveFromFavorites:
        Kita::FavoriteBoards::remove( boardURL );
        break;
    default:
        break;
    }
}

void KitaBoardView::slotMouseButtonClicked( int button, QListViewItem* item )
{
    if ( ! item ) return ;

    QString boardName = item->text( 0 );
    QString boardURL = item->text( 1 );
    if ( boardURL.isEmpty() ) {
        m_boardList->setOpen( item, !item->isOpen() );
        return ;
    }

    emit sigShowSubject();

    switch ( button ) {
    case MidButton:
        emit openURLRequestExt( boardURL, KParts::URLArgs(), "kita_open_2chboard", 1 );
        break;
    case LeftButton:
        if ( KitaConfig::alwaysUseTab() ) {
            emit openURLRequestExt( boardURL, KParts::URLArgs(), "kita_open_2chboard", 1 );
        } else {
            emit openURLRequestExt( boardURL, KParts::URLArgs(), "kita_open_2chboard", 0 );
        }
        break;
    }
}

void KitaBoardView::loadOpened()
{
    QString configPath = locateLocal( "appdata", "board_state.conf" );
    KConfig config( configPath );

    QStringList openedList = config.readListEntry( "Opened" );

    QListViewItem* item;
    for ( item = m_boardList->firstChild(); item; item = item->nextSibling() ) {
        QString categoryName = item->text( 0 );
        if ( openedList.findIndex( categoryName ) != -1 ) {
            item->setOpen( true );
        }
    }
}

void KitaBoardView::saveOpened()
{
    QStringList openedList;
    QListViewItem* item;

    for ( item = m_boardList->firstChild(); item; item = item->nextSibling() ) {
        QString categoryName = item->text( 0 );
        if ( item->isOpen() ) {
            openedList << categoryName;
        }
    }

    QString configPath = locateLocal( "appdata", "board_state.conf" );
    KConfig config( configPath );

    config.writeEntry( "Opened", openedList );
}
