/**************************************************************************
*   Copyright (C) 2003 by Hideki Ikemoto , (c)2004 by 421                 *
*   ikemo@wakaba.jp                                                       *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
**************************************************************************/

/* This class manages the DOM tree  */

#include <kurl.h>
#include <dom/html_element.h>
#include <dom/dom_text.h>
#include <qstringlist.h>

#include "kitadomtree.h"
#include "libkita/kitaconfig.h"
#include "libkita/datmanager.h"
#include "libkita/datinfo.h"
#include "libkita/parsemisc.h"
#include "libkita/kita-utf8.h"


/*---------------------------------------------------------------*/
/*---------------------------------------------------------------*/


KitaDomTree::KitaDomTree( const DOM::HTMLDocument& hdoc, const KURL& datURL )
{
    /* config */

    m_templateNum = 20; /* default number of templates */

    /*-----------------------------*/

    m_hdoc = hdoc;
    m_bufLng = 0;
    m_topNum = 65535;
    m_bottomNum = 0;

    /* create the nodes of footer, header, etc. */
    createKokoyon();
    createFooterOrHeader( "footer" );
    createFooterOrHeader( "header" );
    createTugi100();
    createMae100();

    /* get pointer of DatInfo */
    /* Note that m_datURL is already locked in the KitaHTMLPart. */
    KURL tmpURL = Kita::ParseMisc::parseURLonly( datURL );
    m_datInfo = Kita::DatManager::getDatInfoPointer( tmpURL );
}


KitaDomTree::~KitaDomTree() {}


/* DOM parsing function                       */
/*
  This function parses the res and gets the DOM
  element by DatInfo::getDomElement.
  
  This function does only parsing.
  To show the res, call appendRes later.       */ /* public */

bool KitaDomTree::parseRes( int num )
{
    const int mrg = 50;
    const int delta = 500;

    /*---------------------------------*/

    if ( ! m_datInfo ) return FALSE;
    if ( m_hdoc == NULL ) return FALSE;

    if ( num <= 0 ) return FALSE;

    if ( num < m_bufLng ) {
        if ( m_resparsed[ num ] ) return TRUE; /* already parsed */
    }
    /* resize buffer size */
    else while ( num >= m_bufLng ) {

            if ( m_bufLng == 0 ) m_bufLng = mrg;

            m_bufLng += delta;
            m_res.resize( m_bufLng );
            m_resparsed.resize( m_bufLng, FALSE );
            m_resshown.resize( m_bufLng, FALSE );
        }

    /* parse */
    if ( !m_datInfo->getDomElement( num, m_hdoc, m_res[ num ] ) ) return FALSE;
    m_resparsed[ num ] = TRUE;

    return TRUE;
}



/* append the response */ /* public */
bool KitaDomTree::appendRes( int num )
{
    if ( !parseRes( num ) ) return FALSE;

    m_hdoc.body().appendChild( m_res[ num ] );
    m_resshown[ num ] = TRUE;

    if ( num < m_topNum ) m_topNum = num;
    if ( num > m_bottomNum ) m_bottomNum = num;
    if ( m_topNum <= m_templateNum ) m_topNum = 1;

    return TRUE;
}


/* remove the res */ /* public */
bool KitaDomTree::removeRes( int num )
{
    if ( !parseRes( num ) ) return FALSE;
    if ( !m_resshown[ num ] ) return TRUE;

    m_hdoc.body().removeChild( m_res[ num ] );
    m_resshown[ num ] = FALSE;

    return TRUE;
}



/* redraw all */ /* public */
void KitaDomTree::redraw( bool force )
{

    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;
    m_datInfo->resetAbone();

    int readNum = m_datInfo->getReadNum();

    for ( int i = 1; i <= readNum; i++ ) {

        if ( !m_resparsed[ i ] ) continue;

        QString oldStatus = m_res[ i ].getAttribute( "kita_status" ).string();
        bool abone = m_datInfo->checkAbone( i );
        bool broken = m_datInfo->isResBroken( i );

        /* reparsing */
        if ( force
                || ( oldStatus == "normal" && ( abone || broken ) )
                || ( oldStatus == "abone" && !abone )
                || ( oldStatus == "broken" && !broken )
           ) {

            DOM::Element newNode;
            m_datInfo->getDomElement( i , m_hdoc, newNode );
            if ( m_resshown[ i ] ) m_hdoc.body().replaceChild( newNode, m_res[ i ] );

            /* I don't know how to release the instance of DOM::Element explicitly... */
            m_res[ i ] = newNode;
        }
    }
}


/*-----------------------------------------------------------------*/

/* public information */

/* public */
const bool KitaDomTree::isResShown( int num ) const
{
    return m_resshown[ QMIN( num, m_bufLng - 1 ) ];
}

/* public */
const int KitaDomTree::getTopResNumber() const
{
    return m_topNum;
}

/* public */
const int KitaDomTree::getBottomResNumber() const
{
    return m_bottomNum;
}

/* public */
const int KitaDomTree::getTemplateNumber() const
{
    return m_templateNum;
}

/* public */
const bool KitaDomTree::isMae100Shown() const
{
    if ( m_topNum <= m_templateNum ) return FALSE;

    return TRUE;
}



/*---------------------------------------------------------------------------*/
/* public functions for header Node, footer Node, kokomadeyonda Node, etc... */


/* append templates at the head of thread */ /* public */
void KitaDomTree::appendTemplate( int tmpNum )
{
    if ( tmpNum > m_templateNum ) m_templateNum = tmpNum;

    for ( int i = 1; i <= m_templateNum; i++ ) {
        if ( !parseRes( i ) ) continue;
        if ( m_resshown[ i ] ) continue;
        m_hdoc.body().appendChild( m_res[ i ] );
        m_resshown[ i ] = TRUE;
    }

    if ( m_topNum <= m_templateNum ) m_topNum = 1;
}


/* append footer & header  */ /* public */
void KitaDomTree::appendFooterAndHeader()
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;

    int readNum = m_datInfo->getReadNum();
    if ( !readNum ) return ;

    if ( readNum > m_headerMaxNum + 100 ) {
        updateFooterOrHeader( m_header, "header" );
        updateFooterOrHeader( m_footer, "footer" );
    }

    m_hdoc.body().insertBefore( m_header, m_hdoc.body().firstChild() );
    m_hdoc.body().appendChild( m_footer );
}



/* append kokomadeyonda & tugi 100 & mae 100*/ /* public */
void KitaDomTree::appendKokoyon()
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;

    int readNum = m_datInfo->getReadNum();
    if ( !readNum ) return ;

    int kokoyonNum = m_datInfo->getKokoyonNum();
    if ( kokoyonNum == 0
            || ( kokoyonNum > m_templateNum && kokoyonNum < m_topNum )
            || kokoyonNum > m_bottomNum ) return ;

    int i = kokoyonNum + 1;
    while ( !m_resshown[ i ] && i <= readNum ) i++;

    if ( i <= readNum ) m_hdoc.body().insertBefore( m_kokoyon, m_res[ i ] );
    else m_hdoc.body().appendChild( m_kokoyon );
}


/* public */
void KitaDomTree::appendTugi100()
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;

    int readNum = m_datInfo->getReadNum();
    if ( !readNum ) return ;

    if ( m_bottomNum == readNum ) {
        removeTugi100();
        return ;
    }

    m_hdoc.body().appendChild( m_tugi100 );
}

/* public */
void KitaDomTree::removeTugi100()
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;

    if ( m_tugi100.parentNode() == m_hdoc.body() ) m_hdoc.body().removeChild( m_tugi100 );
}


/* public */
void KitaDomTree::appendMae100()
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;

    int readNum = m_datInfo->getReadNum();
    if ( !readNum ) return ;

    if ( m_topNum <= m_templateNum ) {
        removeMae100();
        return ;
    }

    int i = m_topNum;
    while ( !m_resshown[ i ] && i <= readNum ) i++;

    if ( i <= readNum ) {
        m_hdoc.body().insertBefore( m_mae100, m_res[ i ] );
    } else {
        m_hdoc.body().appendChild( m_mae100 );
    }
}

/* public */
void KitaDomTree::removeMae100()
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;

    if ( !m_mae100.parentNode().isNull() ) m_mae100.parentNode().removeChild( m_mae100 );
}


/* create & append comment node */ /* public */
void KitaDomTree::createCommentNode( const QString& comment, const QString& idstr,
                                     int beforeBR,
                                     int afterBR,
                                     bool binsert )
{
    if ( m_hdoc == NULL ) return ;

    DOM::Element rootnode;

    rootnode = m_hdoc.createElement( "DIV" );
    {
        for ( int i = 0; i < beforeBR; i++ )
            rootnode.appendChild( m_hdoc.createElement( "BR" ) );
        rootnode.setAttribute( "kita_type", "comment" );
        rootnode.setAttribute( "id", idstr );
        rootnode.appendChild( m_hdoc.createTextNode( comment ) );
        for ( int i = 0; i < afterBR; i++ )
            rootnode.appendChild( m_hdoc.createElement( "BR" ) );
    }

    if ( binsert ) {
        if ( m_hdoc.body().firstChild() != NULL ) {
            m_hdoc.body().insertBefore( rootnode, m_hdoc.body().firstChild() );
            return ;
        }
    }

    m_hdoc.body().appendChild( rootnode );
}


/* create belt node */ /* public */
void KitaDomTree::createBeltNode( const QString& idstr )
{
    DOM::Element rootnode;

    QString style = "background-color: #CCCCCC; text-align: center";

    rootnode = m_hdoc.createElement( "DIV" );
    {
        rootnode.setAttribute( "style", style );
        rootnode.setAttribute( "id", idstr );
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );
    }

    m_hdoc.body().appendChild( rootnode );
}


/*--------------------------------------------------------*/

/* private functions */


/* append "A" Node to rootnode */ /* private */
void KitaDomTree::appendANode( DOM::Element rootnode, const QString& linkstr, const QString& comment )
{
    if ( m_hdoc == NULL ) return ;

    DOM::Element tmpelm;
    tmpelm = rootnode.appendChild( m_hdoc.createElement( "A" ) );
    {
        tmpelm.setAttribute( "href", linkstr );
        tmpelm.appendChild( m_hdoc.createTextNode( comment ) );
    }
}


/* update footer or header No  */ /* private */
void KitaDomTree::updateFooterOrHeader(
    DOM::Element& targetelm,   /* = m_header or m_footer */
    QString strType /* = "header" or "footer" */
)
{
    if ( ! m_datInfo ) return ;
    if ( m_hdoc == NULL ) return ;
    if ( strType != "header" && strType != "footer" ) return ;

    DOM::Element bkupElm1, bkupElm2, bkupElm3;
    int readNum = m_datInfo->getReadNum();

    /* remove < a href="#tosaigo"> and <BR> */
    if ( strType == "header" ) {
        bkupElm1 = targetelm.removeChild( targetelm.lastChild() ); /* BR */
        bkupElm2 = targetelm.removeChild( targetelm.lastChild() ); /* BR */
        bkupElm3 = targetelm.removeChild( targetelm.lastChild() ); /* "#tosaigo" */
    } else { /* "footer" */
        bkupElm1 = targetelm.removeChild( targetelm.lastChild() ); /* "#tosaigo" */
    }

    DOM::Node node = targetelm.firstChild(); /* node is now "#kokomade_yonda" */
    node = node.nextSibling(); /* " " */
    node = node.nextSibling(); /* "#zenbu" */
    node = node.nextSibling(); /* " " */
    node = node.nextSibling();

    /* append new "A" nodes */
    for ( int num = 1; num < readNum ; num += 100 ) {

        if ( node == NULL ) {
            QString linkstr = QString( "#%1" ).arg( num );
            appendANode( targetelm, linkstr, QString().setNum( num ) + "-" );
            node = targetelm.appendChild( m_hdoc.createTextNode( " " ) );
            node = node.nextSibling();

            m_headerMaxNum = num;
        } else {
            node = node.nextSibling();
            if ( node != NULL ) node = node.nextSibling();
        }
    }

    /* append < a href="#tosaigo"> and <BR> */
    if ( strType == "header" ) {
        targetelm.appendChild( bkupElm3 ); /* "#tosaigo" */
        targetelm.appendChild( bkupElm2 ); /* BR */
        targetelm.appendChild( bkupElm1 ); /* BR */
    } else { /* "footer" */
        targetelm.appendChild( bkupElm1 ); /* "#tosaigo" */
    }
}


/* create footer or header node   */ /* private */
void KitaDomTree::createFooterOrHeader(
    QString strType /* = "header" or "footer" */
)
{
    if ( m_hdoc == NULL ) return ;
    if ( strType != "header" && strType != "footer" ) return ;

    QString str;
    DOM::Element rootnode, tmpelm;

    rootnode = m_hdoc.createElement( "DIV" );
    {
        rootnode.setAttribute( "kita_type", strType );
        rootnode.setAttribute( "id", strType );

        str = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_KOKOYON );
        appendANode( rootnode, "#kokomade_yonda", str );
        rootnode.appendChild( m_hdoc.createTextNode( " " ) );

        str = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_ZENBU );
        appendANode( rootnode, "#zenbu", str );
        rootnode.appendChild( m_hdoc.createTextNode( " " ) );

        str = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_SAIGO );
        appendANode( rootnode, "#tosaigo", str );

        if ( strType == "header" ) {
            rootnode.appendChild( m_hdoc.createElement( "BR" ) );
            rootnode.appendChild( m_hdoc.createElement( "BR" ) );
        }
    }

    if ( strType == "header" ) m_header = rootnode;
    else m_footer = rootnode;

    m_headerMaxNum = 0;
}


/* create kokomadeyonda node */ /* private */
void KitaDomTree::createKokoyon()
{
    QString str, style;
    DOM::Element rootnode;

    str = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_KOKOYON2 );

    rootnode = m_hdoc.createElement( "DIV" );
    {
        rootnode.setAttribute( "class", "kokoyon" );
        rootnode.setAttribute( "kita_type", "kokoyon" );
        rootnode.setAttribute( "id", "kokomade_yonda" );
        rootnode.appendChild( m_hdoc.createTextNode( str ) );
    }

    m_kokoyon = rootnode;
}



/* create tugi 100 node  */ /* private */
void KitaDomTree::createTugi100()
{
    QString str, style;
    DOM::Element rootnode;

    style = "text-align: center";

    rootnode = m_hdoc.createElement( "DIV" );
    {
        rootnode.setAttribute( "kita_type", "tugi100" );
        rootnode.setAttribute( "id", "tugi100" );
        rootnode.setAttribute( "style", style );

        str = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_TUGI100 );
        appendANode( rootnode, "#tugi100", str );

        rootnode.appendChild( m_hdoc.createTextNode( "  " ) );

        str = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_ZENBUNOKORI );
        appendANode( rootnode, "#nokori", str );

        rootnode.appendChild( m_hdoc.createElement( "BR" ) );
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );
    }

    m_tugi100 = rootnode;
}



/* create mae 100 (template)  node */ /* mae 100 */
void KitaDomTree::createMae100()
{
    QString style;
    DOM::Element rootnode;

    style = "text-align: center";

    rootnode = m_hdoc.createElement( "DIV" );
    {
        rootnode.setAttribute( "kita_type", "mae100" );
        rootnode.setAttribute( "id", "mae100" );
        rootnode.setAttribute( "style", style );

        /* 100 */
        QString tugi100 = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_TUGI100 );
        appendANode( rootnode, "#tmp100", tugi100 );

        /* */
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );

        /* */
        style = "background-color: #CCCCCC; text-align: center";
        DOM::Element divElement = rootnode.appendChild( m_hdoc.createElement( "DIV" ) );
        {
            divElement.setAttribute( "style", style );
            QString templateStr = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_TEMPLATE );
            divElement.appendChild( m_hdoc.createTextNode( templateStr ) );
        }

        /* */
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );

        /* 100 */
        QString mae100 = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_MAE100 );
        appendANode( rootnode, "#mae100", mae100 );

        /* */
        rootnode.appendChild( m_hdoc.createTextNode( "  " ) );

        /*  */
        QString zenbumae = Kita::ParseMisc::utf8ToUnicode( KITAUTF8_ZENBUMAE );
        appendANode( rootnode, "#maezenbu", zenbumae );

        /* */
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );
        rootnode.appendChild( m_hdoc.createElement( "BR" ) );
    }

    m_mae100 = rootnode;
}
